package mvd.jester.simulator;

import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import org.jgrapht.Graphs;
import mvd.jester.model.DagTask;
import mvd.jester.model.Subtask;
import mvd.jester.priority.PriorityManager;
import mvd.jester.simulator.events.*;
import mvd.jester.simulator.exceptions.SchedulingException;
import mvd.jester.simulator.model.ExtendedSubtaskContext;
import mvd.jester.simulator.model.JobContext;
import mvd.jester.simulator.model.ProcessorContext;

public class GlobalScheduler {

    private final int nbProcessors;
    private final PriorityManager priorityManager;

    public GlobalScheduler(PriorityManager priorityManager, int nbProcessors) {
        this.priorityManager = priorityManager;
        this.nbProcessors = nbProcessors;
    }

    public void schedule(Set<DagTask> taskSet, long duration) {
        SchedulingContext sc = new SchedulingContext(priorityManager, nbProcessors, duration);
        taskSet.forEach(t -> sc.events.offer(new TaskReleasedEvent(0, t)));
        while (!sc.events.isEmpty()) {
            IEvent event = sc.events.poll();
            updateExecution(sc, event.getTime());
            handle(event, sc);
        }
    }

    private void handle(IEvent event, SchedulingContext sc) {
        event.log();
        switch (event.getType()) {
            case SUBTASK_ACTIVATED:
                handleSubtaskActivated((SubtaskActivatedEvent) event, sc);
                break;
            case SUBTASK_COMPLETED:
                handleSubtaskCompleted((SubtaskCompletedEvent) event, sc);
                break;
            case SUBTASK_CHECK_COMPLETED:
                handleSubtaskCheckCompleted((SubtaskCheckCompletedEvent) event, sc);
                break;
            case SUBTASK_PREEMPTED:
                handleSubtaskPreempted((SubtaskPreemptedEvent) event, sc);
                break;
            case SUBTASK_STARTED:
                handleSubtaskStarted((SubtaskStartedEvent) event, sc);
                break;
            case SUBTASK_STOPPED:
                handleSubtaskStopped((SubtaskStoppedEvent) event, sc);
                break;
            case JOB_ACTIVATED:
                handleJobActivated((JobActivatedEvent) event, sc);
                break;
            case TASK_RELEASED:
                handleTaskReleased((TaskReleasedEvent) event, sc);
                break;
            default:
                throw new IllegalArgumentException(
                        event.getType() + " is not handled by " + this.getClass());
        }
    }

    private void handleSubtaskActivated(SubtaskActivatedEvent event, SchedulingContext sc) {
        Subtask subtask = event.getSubtask();
        JobContext job = event.getJobContext();
        sc.waitingQueue.offer(new ExtendedSubtaskContext(job, subtask));
        updateAdmission(sc, event.getTime());
    }

    private void handleSubtaskCheckCompleted(SubtaskCheckCompletedEvent event,
            SchedulingContext sc) {
        ProcessorContext processor = event.getProcessor();
        if (processor.isIdle() || !processor.getSubtask().get().equals(event.getSubtaskContext())) {
            throw new SchedulingException(
                    processor + " should execute " + event.getSubtaskContext());
        }
        Optional<ExtendedSubtaskContext> context = processor.getSubtask();
        if (context.isPresent() && context.get().getRemainingExecutionTime() == 0) {
            sc.events.offer(new SubtaskStoppedEvent(event.getTime(), context.get(), processor));
        }
    }

    private void handleSubtaskCompleted(SubtaskCompletedEvent event, SchedulingContext sc) {
        ExtendedSubtaskContext subtask = event.getSubtaskContext();
        JobContext job = subtask.getJobContext();

        DagTask task = subtask.getJobContext().getTask();
        List<Subtask> successor = Graphs.successorListOf(task.getJobDag(), subtask.getSubtask());
        if (successor.isEmpty()) {
            sc.events.offer(new JobCompletedEvent(event.getTime(), job));
            updateAdmission(sc, event.getTime());
        } else {
            for (Subtask s : successor) {
                sc.events.offer(new SubtaskActivatedEvent(event.getTime(), job, s));
            }
        }
    }

    private void handleSubtaskPreempted(SubtaskPreemptedEvent event, SchedulingContext sc) {
        ProcessorContext processor = event.getProcessor();
        long time = event.getTime();
        sc.events.offer(new SubtaskStoppedEvent(time, event.getNextSubtask(), processor));
        sc.events.offer(new SubtaskStartedEvent(time, event.getSubtaskContext(), processor));
    }

    private void handleSubtaskStarted(SubtaskStartedEvent event, SchedulingContext sc) {
        long time = event.getTime();
        ProcessorContext processor = event.getProcessor();
        ExtendedSubtaskContext context = event.getSubtaskContext();
        if (processor.accept(context, time)) {
            sc.events.offer(new SubtaskCheckCompletedEvent(
                    time + context.getRemainingExecutionTime(), context, processor));
        }
    }

    private void handleSubtaskStopped(SubtaskStoppedEvent event, SchedulingContext sc) {
        ExtendedSubtaskContext context = event.getSubtaskContext();
        ProcessorContext processor = event.getProcessor();
        if (context.getRemainingExecutionTime() > 0) {
            sc.waitingQueue.offer(context);
        } else {
            sc.events.offer(new SubtaskCompletedEvent(event.getTime(), context));
        }
        processor.free();
    }

    private void handleJobActivated(JobActivatedEvent event, SchedulingContext sc) {
        JobContext job = event.getJobContext();

        DagTask task = job.getTask();
        long nextRelease = event.getTime() + task.getPeriod();
        if (nextRelease < sc.duration) {
            JobContext nextJob = new JobContext(task, nextRelease);
            sc.events.offer(new JobActivatedEvent(nextRelease, nextJob));
        }

        sc.events.offer(new SubtaskActivatedEvent(event.getTime(), job, job.getSource()));
    }

    private void handleTaskReleased(TaskReleasedEvent event, SchedulingContext sc) {
        DagTask task = event.getTask();
        long release = event.getTime();
        JobContext job = new JobContext(task, release);
        sc.events.offer(new JobActivatedEvent(release, job));
    }

    private void updateAdmission(SchedulingContext sc, long time) {
        for (ProcessorContext processor : sc.processors) {
            if (processor.canAccept(sc.waitingQueue.peek(), priorityManager)) {
                ExtendedSubtaskContext context = sc.waitingQueue.poll();
                if (processor.isIdle()) {
                    sc.events.offer(new SubtaskStartedEvent(time, context, processor));
                } else {
                    sc.events.offer(new SubtaskPreemptedEvent(time, processor.getSubtask().get(),
                            processor, context));
                }
            }
        }
    }

    private void updateExecution(SchedulingContext sc, long time) {
        for (ProcessorContext processor : sc.processors) {
            processor.update(time);
        }
    }

    private class SchedulingContext {

        private final EventQueue events;
        private final Set<ProcessorContext> processors;
        private final WaitQueue waitingQueue;
        private final long duration;

        private SchedulingContext(PriorityManager priorityManager, int nbProcessors,
                long duration) {
            this.events = new EventQueue(priorityManager);
            this.waitingQueue = new WaitQueue(priorityManager);
            this.processors = new HashSet<>();
            for (int m = 0; m < nbProcessors; ++m) {
                processors.add(new ProcessorContext(m));
            }
            this.duration = duration;
        }

    }
}
